#!/usr/bin/env python3

# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

# -*- mode: python-mode -*-

"""List the symbols used to build a goto binary."""

import argparse
import sys

from cbmc_viewer import optionst
from cbmc_viewer import symbolt

################################################################

def create_parser():
    """Command line parser."""

    parser = argparse.ArgumentParser(
        description="List the symbols used to build a goto binary."
    )

    optionst.viewer_symbol(parser)
    optionst.viewer_source(parser)

    optionst.goto(parser)
    optionst.wkdir(parser)
    optionst.srcdir(parser)

    parser.add_argument(
        '--files',
        nargs='+',
        metavar='FILE',
        help="""
        A list of FILES to scan for symbols.
        The files may be specified as either absolute paths or paths
        relative to root.
        """
    )

    optionst.log(parser)

    return parser

################################################################

def main():
    """List the symbols used to build a goto binary."""

    args = create_parser().parse_args()
    args = optionst.defaults(args)

    try:
        symbols = symbolt.make_symbol(args.viewer_symbol, args.viewer_source,
                                      args.goto,
                                      args.wkdir,
                                      args.srcdir, args.files)
        print(symbols)
    except UserWarning as error:
        sys.exit(error)

if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

# -*- mode: python-mode -*-

"""List CBMC coverage checking results."""


import argparse
import sys

from cbmc_viewer import coveraget
from cbmc_viewer import optionst

def create_parser():
    """Command line parser."""

    parser = argparse.ArgumentParser(
        description="List CBMC coverage checking results."
    )

    optionst.viewer_coverage(parser)
    optionst.srcdir(parser)

    parser.add_argument(
        'cbmc_coverage',
        nargs='*',
        help="""
        An xml or json file containing the output of 'cbmc --location cover'.
        Merge the coverage data if more than one file is given.
        Do not mix xml and json files.
        """
    )

    optionst.log(parser)

    return parser

################################################################

def main():
    """List CBMC coverage checking results."""

    args = create_parser().parse_args()
    args = optionst.defaults(args)

    try:
        coverage = coveraget.make_coverage(args.viewer_coverage, args.srcdir, args.cbmc_coverage)
        print(coverage)
    except UserWarning as error:
        sys.exit(error)

if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

# -*- mode: python-mode -*-

"""List CBMC property checking results."""

import argparse
import sys

from cbmc_viewer import optionst
from cbmc_viewer import resultt

def create_parser():
    """Command line parser."""

    parser = argparse.ArgumentParser(
        description='List CBMC property checking results.'
    )

    optionst.viewer_result(parser)
    parser.add_argument(
        'cbmc_result',
        nargs='*',
        help="""
        One or more files containing the output of cbmc property checking
        as text, xml, or json.
        Multiple files will be merged into a single list of
        property checking results.
        Do not mix text, xml, and json files.
        """
    )

    optionst.log(parser)

    return parser

################################################################

def main():
    """List CBMC property checking results."""

    args = create_parser().parse_args()
    args = optionst.defaults(args)

    try:
        results = resultt.make_result(args.viewer_result, args.cbmc_result)
        print(results)
    except UserWarning as error:
        sys.exit(error)

if __name__ == "__main__":
    main()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""Parsing of xml and json files with limited error handling."""

from xml.etree import ElementTree

import json
import logging
import re

def parse_xml_file(xfile, fail=False):
    """Open and parse an xml file."""

    try:
        with open(xfile) as data:
            return parse_xml_string(data.read(), xfile, fail)
    except IOError as err:
        message = "Can't open xml file {}: {}".format(xfile, err.strerror)
        logging.info(message)
        if fail:
            raise UserWarning(message) from None
    return None

def parse_xml_string(xstr, xfile=None, fail=False):
    """Parse an xml string."""

    try:
        # Messages printed by cbmc before the coverage data we care
        # about may quote from the code, and quoted null character
        # '\0' will appear as the string '&#0', which is unparsable by
        # ElementTree.
        xstr = xstr.replace('&#0;', 'null_char')
        return ElementTree.fromstring(xstr)
    except ElementTree.ParseError as err:
        if xfile:
            message = "Can't parse xml file {}: string {}...: {}".format(
                xfile, xstr[:40], str(err)
            )
        else:
            message = "Can't parse xml string {}...: {}".format(
                xstr[:40], str(err)
            )
        logging.info(message)
        if fail:
            raise UserWarning(message) from None
    return None

def parse_json_file(jfile, fail=False, goto_analyzer=False):
    """Open and parse an json file."""

    try:
        with open(jfile) as data:
            return parse_json_string(data.read(), jfile, fail, goto_analyzer)
    except IOError as err:
        message = "Can't open json file {}: {}".format(jfile, err.strerror)
        logging.info(message)
        if fail:
            raise UserWarning(message) from None
    return None

def parse_json_string(jstr, jfile=None, fail=False, goto_analyzer=False):
    """Parse an json string."""

    try:
        jstr = clean_up_goto_analyzer(jstr) if goto_analyzer else jstr
        return json.loads(jstr)
    except json.JSONDecodeError as err:
        if jfile:
            message = "Can't parse json file {}: string {}...: {}".format(
                jfile, jstr[:40], str(err)
            )
        else:
            message = "Can't parse json string {}...: {}".format(
                jstr[:40], str(err)
            )
        logging.info(message)
        if fail:
            raise UserWarning(message) from None

    return None

def clean_up_goto_analyzer(json_data):
    """Clean up the json output of goto-analyzer."""

    # the json produced by goto-analyzer is preceeded by a block of text
    json_data = re.sub(r'^.*\n\[', '[', json_data, flags=re.DOTALL)
    # the json produced by goto-analyzer sometimes omits line numbers
    json_data = json_data.replace('"first line": ,\n', '"first line": 0,\n')
    json_data = json_data.replace('"last line": \n', '"last line": 0\n')
    return json_data



#!/usr/bin/env python3

# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

# -*- mode: python-mode -*-

"""List the source files used to build a goto binary."""

import argparse
import sys

from cbmc_viewer import optionst
from cbmc_viewer import sourcet

################################################################

def create_parser():
    """Command line parser."""

    parser = argparse.ArgumentParser(
        description='List the source files used to build a goto binary.')

    optionst.viewer_source(parser)
    optionst.goto(parser)
    optionst.srcdir(parser)
    optionst.wkdir(parser)
    optionst.source_method(parser)
    optionst.extensions(parser)
    optionst.exclude(parser)
    optionst.log(parser)

    return parser

################################################################

def main():
    """List the source files used to build a goto binary."""

    args = create_parser().parse_args()
    args = optionst.defaults(args)

    try:
        sources = sourcet.make_source(args.viewer_source,
                                      args.goto,
                                      args.source_method,
                                      args.srcdir, args.wkdir,
                                      args.exclude, args.extensions)
        print(sources)
    except UserWarning as error:
        sys.exit(error)

################################################################

if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

# -*- mode: python-mode -*-

"""List the loops in a goto binary."""

import argparse
import sys

from cbmc_viewer import loopt
from cbmc_viewer import optionst

################################################################

def create_parser():
    """Command line parser."""

    parser = argparse.ArgumentParser(
        description='List loops in a goto binary.'
    )

    optionst.viewer_loop(parser)
    optionst.goto(parser)
    optionst.srcdir(parser)

    parser.add_argument(
        'cbmc_loop',
        nargs='*',
        help="""
        A list of xml or json files containing the output of
        'cbmc --show-loops'.  Do not mix xml and json files.
        """
    )

    optionst.log(parser)

    return parser

################################################################

def main():
    """List the loops in a goto binary."""

    args = create_parser().parse_args()
    args = optionst.defaults(args)

    try:
        loops = loopt.make_loop(args.viewer_loop, args.cbmc_loop, args.srcdir, args.goto)
        print(loops)
    except UserWarning as error:
        sys.exit(error)

if __name__ == "__main__":
    main()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""Command line options."""

import logging
import os
import platform

from cbmc_viewer import filet
from cbmc_viewer import version as viewer_version
from cbmc_viewer.sourcet import Sources

def goto(parser):
    'Define --goto command line option.'

    parser.add_argument(
        '--goto',
        help="""
        The goto binary.
        """
    )
    return parser

def srcdir(parser):
    'Define --srcdir command line option.'

    parser.add_argument(
        '--srcdir',
        type=os.path.abspath,
        help="""
        The source directory.  The root of the source tree.
        """
    )
    return parser

def wkdir(parser):
    'Define --wkdir command line option.'

    parser.add_argument(
        '--wkdir',
        default=".",
        type=os.path.abspath,
        help="""
        The working directory in source locations in the goto binary
        (and omitted when source locations are printed in textual form).
        This is usually the directory in which goto-cc was invoked to
        build the goto binary.
        """
    )
    return parser

def reportdir(parser):
    'Define --reportdir command line option.'

    parser.add_argument(
        '--reportdir',
        default='report',
        type=os.path.abspath,
        help="""
        The report directory.  Write the final report to this directory.
        (Default: %(default)s)
        """
    )
    return parser

def result(parser):
    'Define --result command line option.'

    parser.add_argument(
        '--result',
        metavar='FILE',
        nargs='+',
        help="""
        CBMC property checking results.
        A text, xml, or json file containing the output of 'cbmc'.
        """
    )
    return parser

def coverage(parser):
    'Define --coverage command line option.'

    parser.add_argument(
        '--coverage',
        metavar='FILE',
        nargs='+',
        help="""
        CBMC coverage checking results.
        An xml or json file containing the output of
        'cbmc --cover locations'.
        """
    )
    return parser

def property(parser):
    'Define --property command line option.'

    # pylint: disable=redefined-builtin

    parser.add_argument(
        '--property',
        metavar='FILE',
        nargs='+',
        help="""
        CBMC properties checked during property checking.
        An xml or json file containing the output of
        'cbmc --show-properties'.
        """
    )
    return parser

def exclude(parser):
    'Define --exclude command line option.'

    parser.add_argument(
        '--exclude',
        help="""
        Paths relative to SRCDIR to exclude from the list of source files.
        A Python regular expression matched against the result
        of os.path.normpath().  The match is case insensitive.
        """
    )
    return parser

def extensions(parser):
    'Define --extensions command line option.'

    parser.add_argument(
        '--extensions',
        metavar='REGEXP',
        default=r'^\.(c|h|inl)$',
        help="""
        File extensions of files to include in the list of source files.
        A Python regular expression matched against the result of
        os.path.splitext().  The match is case insensitive.
        (Default: %(default)s)
        """
    )
    return parser

def source_method(parser):
    'Define --source-method command line option.'

    choices = []
    if platform.system() != 'Windows':
        choices.append('find')
    choices.extend(['walk', 'make', 'goto'])

    parser.add_argument(
        '--source-method',
        metavar='MHD',
        choices=choices,
        help="""
        The method to use to list source files under SRCDIR.  Methods
        available are [%(choices)s]: Use the Linux 'find' command, use
        the Python 'walk' method, use the 'make' command to build
        the goto binary with the preprocessor, or use the symbol table in
        the goto binary.  The default method is 'goto' if
        SRCDIR and WKDIR and GOTO are specified, 'make' if SRCDIR and WKDIR
        are specified, 'walk' on Windows, and 'find' otherwise.
        """
    )
    return parser

def config(parser):
    'Define --config command line option.'

    parser.add_argument(
        '--config',
        metavar='JSON',
        default="cbmc-viewer.json",
        help="""
        JSON configuration file. (Default: '%(default)s')
        """
    )
    return parser

################################################################
# Load data from the make-* commands

def viewer_reachable(parser):
    "Load make-reachable data"

    parser.add_argument(
        '--viewer-reachable',
        metavar='JSON',
        nargs='+',
        help='Load reachable functions from the JSON output of make-reachable.'
        ' If multiple files are given, merge multiple data sets into one.'
    )

def viewer_coverage(parser):
    "Load make-coverage data"

    parser.add_argument(
        '--viewer-coverage',
        metavar='JSON',
        nargs='+',
        help='Load coverage data from the JSON output of make-coverage.'
        ' If multiple files are given, merge multiple data sets into one.'
    )

def viewer_loop(parser):
    "Load make-loop data"

    parser.add_argument(
        '--viewer-loop',
        metavar='JSON',
        nargs='+',
        help='Load loops from the JSON output of make-loop.'
        ' If multiple files are given, merge multiple data sets into one.'
    )

def viewer_property(parser):
    "Load make-property data"

    parser.add_argument(
        '--viewer-property',
        metavar='JSON',
        nargs='+',
        help='Load properties from the JSON output of make-property.'
        ' If multiple files are given, merge multiple data sets into one.'
    )

def viewer_result(parser):
    "Load make-result data"


    parser.add_argument(
        '--viewer-result',
        metavar='JSON',
        nargs='+',
        help='Load results from the JSON output of make-result.'
        ' If multiple files are given, merge multiple data sets into one.'
    )

def viewer_source(parser):
    "Load make-source data"

    parser.add_argument(
        '--viewer-source',
        metavar='JSON',
        nargs='+',
        help='Load sources from the JSON output of make-source.'
        ' If multiple files are given, merge multiple data sets into one.'
    )
def viewer_symbol(parser):
    "Load make-symbol data"

    parser.add_argument(
        '--viewer-symbol',
        metavar='JSON',
        nargs='+',
        help='Load symbols from the JSON output of make-symbol.'
        ' If multiple files are given, merge multiple data sets into one.'
    )

def viewer_trace(parser):
    "Load make-trace data"

    parser.add_argument(
        '--viewer-trace',
        metavar='JSON',
        nargs='+',
        help='Load traces from the JSON output of make-trace.'
        ' If multiple files are given, merge multiple data sets into one.'
    )
    return parser

################################################################
# Other line options

def log(parser):
    'Define --verbose and --debug command line options.'

    parser.add_argument(
        '--verbose',
        action='store_true',
        help='Verbose output.'
    )
    parser.add_argument(
        '--debug',
        action='store_true',
        help='Debugging output.'
    )
    return parser

def version(parser):
    'Define --version command line option.'

    parser.add_argument(
        '--version',
        action='version',
        version=viewer_version.version(),
        help="""Display version number and exit."""
    )
    return parser

################################################################
# Depricated command line options
#
# Handle command line options from prior versions of cbmc-viewer that
# are deprecated in the current version.  Some old options have
# natural equivalents among the new options, some have a slight change
# in semantics that aren't easy to fix, and some we no longer support.

def block(parser):
    'Define --block command line option.'
    parser.add_argument(
        '--block',
        metavar='BLOCK',
        help="""Use --coverage instead."""
    )
    return parser

def htmldir(parser):
    'Define --htmldir command line option.'
    parser.add_argument(
        '--htmldir',
        metavar='HTMLDIR',
        help="""Use --reportdir instead."""
    )
    return parser

def srcexclude(parser):
    'Define --srcexclude command line option.'
    parser.add_argument(
        '--srcexclude',
        metavar='EXCLUDE',
        help="""Use --exclude instead."""
    )
    return parser

def blddir(parser):
    'Define --blddir command line option.'
    parser.add_argument(
        '--blddir',
        help="""Ignored."""
    )
    return parser

def storm(parser):
    'Define --storm command line option.'
    parser.add_argument(
        '--storm',
        help="""Ignored."""
    )
    return parser

def handle_deprecated_arguments(args):
    """Warn about deprecated arguments, use them  when possible."""

    if hasattr(args, 'block') and args.block:
        if hasattr(args, 'coverage'):
            logging.warning("--block is deprecated, using --coverage %s.",
                            args.block)
            args.coverage = [args.block] # block is a string, coverage is a list
        else:
            logging.warning("--block is deprecated, use --coverage instead.")
        args.block = None

    if hasattr(args, 'htmldir') and args.htmldir:
        if hasattr(args, 'reportdir'):
            logging.warning("--htmldir is deprecated, using --reportdir %s.",
                            args.htmldir)
            args.reportdir = args.htmldir
        else:
            logging.warning("--htmldir is deprecated, use --reportdir instead.")
        args.htmldir = None

    if hasattr(args, 'srcexclude') and args.srcexclude:
        if hasattr(args, 'exclude'):
            logging.warning("--srcexclude is deprecated, using --exclude %s.",
                            args.srcexclude)
            args.exclude = args.srcexclude
        else:
            logging.warning("--srcexclude is deprecated, "
                            "use --exclude instead.")
        logging.warning("--srcexclude and --exclude use slight different "
                        "regular expressions.")
        args.srcexclude = None

    if hasattr(args, 'blddir') and args.blddir:
        logging.warning("--blddir is deprecated, ignoring --blddir.")
        args.blddir = None

    if hasattr(args, 'storm') and args.storm:
        logging.warning("--storm is deprecated, ignoring --storm.")
        args.storm = None

    return args

################################################################
# Set default values for arguments

def default_logging(args):
    'Set default logging configuration.'

    # Only the first invocation of basicConfig configures the root logger
    if getattr(args, 'debug', False):
        logging.basicConfig(level=logging.DEBUG,
                            format='%(levelname)s: %(message)s')
    if getattr(args, 'verbose', False):
        logging.basicConfig(level=logging.INFO,
                            format='%(levelname)s: %(message)s')
    logging.basicConfig(format='%(levelname)s: %(message)s')

    return args

def default_source_method(args):
    'Set default source method.'

    # Set source_method to an enum
    if hasattr(args, 'source_method'):

        # Set source method to its enum value or None
        args.source_method = {
            'find': Sources.FIND,
            'walk': Sources.WALK,
            'make': Sources.MAKE,
            'goto': Sources.GOTO,
            None: None
        }[args.source_method]

        # Set source method to a reasonable default value
        if args.source_method is None:
            if getattr(args, 'srcdir') is not None:
                if getattr(args, 'wkdir') is not None:
                    if getattr(args, 'goto') is not None:
                        args.source_method = Sources.GOTO
                    else:
                        args.source_method = Sources.MAKE
                elif platform.system() == 'Windows':
                    args.source_method = Sources.WALK
                else:
                    args.source_method = Sources.FIND

        # Confirm existence of command line options needed by source method
        if args.source_method == Sources.GOTO:
            if (not hasattr(args, 'srcdir') or
                    not hasattr(args, 'wkdir') or
                    not hasattr(args, 'goto')):
                raise UserWarning('Options --srcdir, --wkdir, and --goto '
                                  'required by source method goto.')

        if args.source_method == Sources.FIND:
            if not hasattr(args, 'srcdir'):
                raise UserWarning('Option --srcdir required '
                                  'by source method find.')

        if args.source_method == Sources.WALK:
            if not hasattr(args, 'srcdir'):
                raise UserWarning('Option --srcdir required '
                                  'by source method walk.')

        if args.source_method == Sources.MAKE:
            if not hasattr(args, 'srcdir') or not hasattr(args, 'wkdir'):
                raise UserWarning('Options --srcdir and --wkdir required '
                                  'by source method make.')

    return args

def warn_against_using_text_for_cbmc_output(args):
    'Recommend the use of xml or json input instead of text.'

    for attr in ['result', 'coverage', 'property']:
        filenames = getattr(args, attr, None)
        if filenames and filet.any_text_files(filenames):
            logging.warning("Use xml or json instead of text for "
                            "better results: %s", filenames)

def defaults(args):
    'Set default values based on command line arguments.'

    args = default_logging(args)
    args = handle_deprecated_arguments(args)
    args = default_source_method(args)
    warn_against_using_text_for_cbmc_output(args)

    if hasattr(args, 'srcdir') and args.srcdir is not None:
        args.srcdir = os.path.abspath(args.srcdir)
    if hasattr(args, 'wkdir') and args.wkdir is not None:
        args.wkdir = os.path.abspath(args.wkdir)

    return args

################################################################



#!/usr/bin/env python3

# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

# -*- mode: python-mode -*-

"""Properties checked by CBMC during property checking."""

import argparse
import sys

from cbmc_viewer import optionst
from cbmc_viewer import propertyt

################################################################

def create_parser():
    """Command line parser."""

    parser = argparse.ArgumentParser(
        description='Properties checked by CBMC during property checking.'
    )

    optionst.viewer_property(parser)
    optionst.srcdir(parser)

    parser.add_argument(
        'cbmc_property',
        nargs='*',
        help="""
        One or more files containing the xml or json output of
        'cbmc --show-properties'.  Do not mix xml and json files.
        """
    )

    optionst.log(parser)

    return parser

################################################################

def main():
    """Properties checked by CBMC during property checking."""

    args = create_parser().parse_args()
    args = optionst.defaults(args)

    try:
        properties = propertyt.make_property(args.viewer_property, args.cbmc_property, args.srcdir)
        print(properties)
    except UserWarning as error:
        sys.exit(error)

if __name__ == "__main__":
    main()
